/* Copyright (c) 2012 imacat
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/* 
 * CalcMosaic
 * 
 * Created on 2012-08-12
 * 
 * Copyright (c) 2012 imacat
 */

package tw.idv.imacat.calcmosaic;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.ResourceBundle;
import javax.imageio.ImageIO;

/**
 * A program to create mosaic art from images with Calc.
 * 
 * @author <a href="mailto:imacat&#64;mail.imacat.idv.tw">imacat</a>
 * @version 1.0.1
 */
public class CalcMosaic {
    
    /** The version number. */
    private final static String VERSION = "1.0.1";
    
    /** The version number. */
    private final static String AUTHOR = "imacat <imacat@mail.imacat.idv.tw>";
    
    /** The version number. */
    private final static String JAR_NAME = "calcmosaic.jar";
    
    /** The module name of the BASIC playing macros. */
    private static final String playModule = "_0Play";
    
    /** The module name of the BASIC macros. */
    private static final String basicModule = "_1CalcMosaic";
    
    /** The BASIC macros to play the mosaic art. */
    private static final String playMacros =
"' Copyright (c) 2012 imacat\n"
+ "' \n"
+ "' Licensed under the Apache License, Version 2.0 (the \"License\");\n"
+ "' you may not use this file except in compliance with the License.\n"
+ "' You may obtain a copy of the License at\n"
+ "' \n"
+ "'     http://www.apache.org/licenses/LICENSE-2.0\n"
+ "' \n"
+ "' Unless required by applicable law or agreed to in writing, software\n"
+ "' distributed under the License is distributed on an \"AS IS\" BASIS,\n"
+ "' WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n"
+ "' See the License for the specific language governing permissions and\n"
+ "' limitations under the License.\n"
+ "\n"
+ "' _0Play: The BASIC macros to play the spreadsheets\n"
+ "'   Created with " + JAR_NAME + " version " + VERSION + "\n"
+ "'   By imacat <imacat@mail.imacat.idv.tw>, 2012/10/5\n"
+ "\n"
+ "Option Explicit\n"
+ "\n"
+ "' The default background music, related to the document file\n"
+ "Const sDefaultMusic = \"\"\n"
+ "' The default break time between sheets (ms)\n"
+ "Const nDefaultBreak = 100\n"
+ "\n"
+ "' subPlaySheets: Plays the spreadsheets.\n"
+ "'   1. Save file first for the macro to work.\n"
+ "'   2. Hit the start button at A1 of the first\n"
+ "'      and last sheet to play the sheets.\n"
+ "'   3. You can set the break time (in\n"
+ "'      milliseconds) between sheets at B1 of\n"
+ "'      the first sheet (default 100 ms).\n"
+ "'   4. You can set the background music at C1\n"
+ "'      of the first sheet.  This only works on\n"
+ "'      Linux and you must install mplayer.\n"
+ "'      Relative file path is relative to the\n"
+ "'      document file.  Play once before your\n"
+ "'      live show to avoid the I/O delay with cache.\n"
+ "Sub subPlaySheets\n"
+ "\tDim oController As Object, oSheets As Object\n"
+ "\tDim nBreak As Integer, nI As Integer\n"
+ "\tDim nCount As Integer, oSheet As Object\n"
+ "\t\n"
+ "\toController = thisComponent.getCurrentController\n"
+ "\toSheets = thisComponent.getSheets\n"
+ "\tnBreak = oSheets.getByIndex (0).getCellByPosition (1, 0).getValue\n"
+ "\tIf nBreak = 0 Then\n"
+ "\t\tnBreak = nDefaultBreak\n"
+ "\tEnd If\n"
+ "\tnCount = oSheets.getCount\n"
+ "\t\n"
+ "\tsubPlayMusic\n"
+ "\t\n"
+ "\tFor nI = 0 To nCount - 1\n"
+ "\t\toSheet = oSheets.getByIndex (nI)\n"
+ "\t\toController.setActiveSheet (oSheet)\n"
+ "\t\tWait nBreak\n"
+ "\tNext nI\n"
+ "End Sub\n"
+ "\n"
+ "' subPlayMusic: Plays the background music.\n"
+ "Sub subPlayMusic\n"
+ "\tDim oSheet As Object, sMusic As String, sFile As String, nPos As Integer\n"
+ "\t\n"
+ "\t' This only works on Linux and you must install mplayer.\n"
+ "\tIf Not FileExists (\"/usr/bin/mplayer\") Then\n"
+ "\t\tExit Sub\n"
+ "\tEnd if\n"
+ "\t\n"
+ "\toSheet = thisComponent.getSheets.getByIndex (0)\n"
+ "\tsMusic = oSheet.getCellByPosition (2, 0).getString\n"
+ "\tIf sMusic = \"\" Then\n"
+ "\t\tsMusic = sDefaultMusic\n"
+ "\tEnd If\n"
+ "\tIf sMusic = \"\" Then\n"
+ "\t\tExit Sub\n"
+ "\tEnd If\n"
+ "\t\n"
+ "\t' Relative file path is relative to the document file.\n"
+ "\tIf Left (sMusic, 1) <> \"/\" Then\n"
+ "\t\tsFile = ConvertFromURL (thisComponent.getLocation)\n"
+ "\t\tIf sFile = \"\" Then\n"
+ "\t\t\tExit Sub\n"
+ "\t\tEnd If\n"
+ "\t\tnPos = Len (sFile)\n"
+ "\t\tDo While  Mid (sFile, nPos, 1) <> \"/\"\n"
+ "\t\t\tnPos = nPos - 1\n"
+ "\t\tLoop\n"
+ "\t\tsMusic = Left (sFile, nPos) & sMusic\n"
+ "\tEnd If\n"
+ "\t\n"
+ "\tShell \"mplayer\", 6, \"\"\"\" & sMusic & \"\"\"\", False\n"
+ "End Sub\n";
    
    /** The BASIC macros to convert color numbers to mosaic art. */
    private static final String basicMacros =
"' Copyright (c) 2012 imacat\n"
+ "' \n"
+ "' Licensed under the Apache License, Version 2.0 (the \"License\");\n"
+ "' you may not use this file except in compliance with the License.\n"
+ "' You may obtain a copy of the License at\n"
+ "' \n"
+ "'     http://www.apache.org/licenses/LICENSE-2.0\n"
+ "' \n"
+ "' Unless required by applicable law or agreed to in writing, software\n"
+ "' distributed under the License is distributed on an \"AS IS\" BASIS,\n"
+ "' WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n"
+ "' See the License for the specific language governing permissions and\n"
+ "' limitations under the License.\n"
+ "\n"
+ "' _1CalcMosaic: The BASIC macros of Calc Mosaic\n"
+ "'   Created with " + JAR_NAME + " version " + VERSION + "\n"
+ "'   By imacat <imacat@mail.imacat.idv.tw>, 2012/10/5\n"
+ "\n"
+ "Option Explicit\n"
+ "\n"
+ "' subCreateMosaic: Creates the spreadsheets of mosaic art\n"
+ "Sub subCreateMosaic (mData () As Object)\n"
+ "\tDim mNames (UBound (mData (0))) As String\n"
+ "\tDim mColors (UBound (mData (1))) As Object\n"
+ "\tDim oSheet As Object, nRows As Integer, nColumns As Integer\n"
+ "\tDim nCurrentRow As Integer, nI As Integer, oRange As Object\n"
+ "\tDim oCell As Object\n"
+ "\t\n"
+ "\t' Adds the spreadsheets and allocates memory first, to prevent\n"
+ "\t' memory corruption when the the number of sheets goes large.\n"
+ "\tmNames = mData (0)\n"
+ "\tmColors = mData (1)\n"
+ "\tsubInitializeSheets (mNames)\n"
+ "\tFor nI = 0 To UBound (mNames)\n"
+ "\t\tsubCreateMosaicSheet (mNames (nI), mColors (nI))\n"
+ "\tNext nI\n"
+ "\t\n"
+ "\tsubAddStartButtons\n"
+ "\toSheet = thisComponent.getSheets.getByIndex (0)\n"
+ "\toCell = oSheet.getCellByPosition (1, 0)\n"
+ "\toCell.setPropertyValue (\"CharColor\", mColors (0)(0)(1))\n"
+ "\toCell = oSheet.getCellByPosition (2, 0)\n"
+ "\toCell.setPropertyValue (\"CharColor\", mColors (0)(0)(2))\n"
+ "\tthisComponent.getCurrentController.setActiveSheet (oSheet)\n"
+ "\tsubShowNotice\n"
+ "End Sub\n"
+ "\n"
+ "' subCreateMosaicSheet: Creates a spreadsheet of mosaic art with\n"
+ "'   the color values\n"
+ "Sub subCreateMosaicSheet (sSheet As String, mColors () As Object)\n"
+ "\tDim oSheet As Object, oRange As Object, oCell As Object\n"
+ "\tDim nRow As Integer, nColumn As Integer, nWidth As Integer\n"
+ "\t\n"
+ "\toSheet = thisComponent.getSheets.getByName (sSheet)\n"
+ "\toRange = oSheet.getCellRangeByPosition ( _\n"
+ "\t\t0, 0, UBound (mColors (0)), UBound (mColors))\n"
+ "\tFor nRow = 0 To UBound (mColors)\n"
+ "\t\tFor nColumn = 0 To UBound (mColors (0))\n"
+ "\t\t\toCell = oRange.getCellByPosition (nColumn, nRow)\n"
+ "\t\t\toCell.setPropertyValue (\"CellBackColor\", _\n"
+ "\t\t\t\tCLng (mColors (nRow)(nColumn)))\n"
+ "\t\tNext nColumn\n"
+ "\tNext nRow\n"
+ "\t\n"
+ "\t' Shortly displays the sheet.  Removing these can save the\n"
+ "\t' display refreshing time and run slightly faster, with\n"
+ "\t' the cost of not seeing the progress and looking like hang.\n"
+ "\tnWidth = oRange.getRows.getPropertyValue (\"Height\")\n"
+ "\toRange.getColumns.setPropertyValue (\"Width\", nWidth)\n"
+ "\tthisComponent.getCurrentController.setActiveSheet (oSheet)\n"
+ "\tWait 1\n"
+ "End Sub\n"
+ "\n"
+ "' subInitializeSheets: Initialize the spreadsheets\n"
+ "Sub subInitializeSheets (mNames () As Object)\n"
+ "\tDim oSheets As Object, nI As Integer, mExisting () As String\n"
+ "\t\n"
+ "\toSheets = thisComponent.getSheets\n"
+ "\tIf oSheets.hasByName (mNames (0)) Then\n"
+ "\t\tIf oSheets.getCount = 1 Then\n"
+ "\t\t\toSheets.insertNewByName (mNames (0) & 1, 0)\n"
+ "\t\tEnd If\n"
+ "\t\toSheets.removeByName (mNames (0))\n"
+ "\tEnd If\n"
+ "\tmExisting = oSheets.getElementNames\n"
+ "\toSheets.insertNewByName (mNames (0), 0)\n"
+ "\tFor nI = 0 To UBound (mExisting)\n"
+ "\t\toSheets.removeByName (mExisting (nI))\n"
+ "\tNext nI\n"
+ "\tFor nI = 1 To UBound (mNames)\n"
+ "\t\toSheets.insertNewByName (mNames (nI), nI)\n"
+ "\tNext nI\n"
+ "End Sub\n"
+ "\n"
+ "' fnGetEmptyStringArray: Returns an empty array of String.\n"
+ "Function fnGetEmptyStringArray (nUBound As Integer)\n"
+ "\tDim mArray (nUBound) As String\n"
+ "\t\n"
+ "\tfnGetEmptyStringArray = mArray\n"
+ "End Function\n"
+ "\n"
+ "' subAddStartButtons: Adds start buttons to show the sheets\n"
+ "Sub subAddStartButtons\n"
+ "\tDim oSheets As Object\n"
+ "\t\n"
+ "\toSheets = thisComponent.getSheets\n"
+ "\tsubAddStartButton (oSheets.getByIndex (0))\n"
+ "\t' Adds a start button at the end.  Last time (2012/8/18 on\n"
+ "\t' COSCUP 2012 lightning talk) I tested playing and ended at\n"
+ "\t' the last sheet.  When talk started I did not notice it and\n"
+ "\t' spent quite some time finding the button.  Audience got\n"
+ "\t' anxious and even lost attention when the show begins.\n" 
+ "\tsubAddStartButton (oSheets.getByIndex (oSheets.getCount - 1))\n"
+ "End Sub\n"
+ "\n"
+ "' subAddStartButton: Adds a start button to show the sheets\n"
+ "Sub subAddStartButton (oSheet As Object)\n"
+ "\tDim oPage As Object, oShape As Object, oButton As Object\n"
+ "\tDim nSize As Integer, nColor As Long\n"
+ "\tDim aPos As New com.sun.star.awt.Point\n"
+ "\tDim aSize As New com.sun.star.awt.Size\n"
+ "\tDim oForm As Object, nIndex As Integer\n"
+ "\tDim oEvent As New com.sun.star.script.ScriptEventDescriptor\n"
+ "\t\n"
+ "\toPage = oSheet.getDrawPage\n"
+ "\tnSize = oSheet.getRows.getPropertyValue (\"Height\")\n"
+ "\t\n"
+ "\toShape = thisComponent.createInstance ( _\n"
+ "\t\t\"com.sun.star.drawing.ControlShape\")\n"
+ "\taPos.X = 0\n"
+ "\taPos.Y = 0\n"
+ "\toShape.setPosition (aPos)\n"
+ "\taSize.Width = nSize\n"
+ "\taSize.Height = nSize\n"
+ "\toShape.setSize (aSize)\n"
+ "\t\n"
+ "\toButton = thisComponent.createInstance ( _\n"
+ "\t\t\"com.sun.star.form.component.CommandButton\")\n"
+ "\tnColor = oSheet.getCellByPosition (0, 0).getPropertyValue ( _\n"
+ "\t\t\"CellBackColor\")\n"
+ "\toButton.setPropertyValue (\"BackgroundColor\", nColor)\n"
+ "\t\n"
+ "\toShape.setControl (oButton)\n"
+ "\toPage.add (oShape)\n"
+ "\t\n"
+ "\toEvent.ListenerType = \"XActionListener\"\n"
+ "\toEvent.EventMethod = \"actionPerformed\"\n"
+ "\toEvent.ScriptType = \"Script\"\n"
+ "\toEvent.ScriptCode = \"vnd.sun.star.script:\" _\n"
+ "\t\t& \"Standard." + playModule + ".subPlaySheets\" _\n"
+ "\t\t& \"?language=Basic&location=document\"\n"
+ "\toForm = oButton.getParent\n"
+ "\tnIndex = oForm.getCount - 1\n"
+ "\toForm.registerScriptEvent (nIndex, oEvent)\n"
+ "End Sub\n"
+ "\n"
+ "' subShowNotice: Shows the notice message.\n"
+ "Sub subShowNotice\n"
+ "\tMsgBox \"Done.  Notice:\" & Chr (13) _\n"
+ "\t\t& \"1. Save file first for the macro to work.\" & Chr (13) _\n"
+ "\t\t& \"2. Hit the start button at A1 of the first\" & Chr (13) _\n"
+ "\t\t& \"   and last sheet to play the sheets.\" & Chr (13) _\n"
+ "\t\t& \"3. You can set the break time (in\" & Chr (13) _\n"
+ "\t\t& \"   milliseconds) between sheets at B1 of\" & Chr (13) _\n"
+ "\t\t& \"   the first sheet (default 100 ms).\" & Chr (13) _\n"
+ "\t\t& \"4. You can set the background music at C1\" & Chr (13) _\n"
+ "\t\t& \"   of the first sheet.  This only works on\" & Chr (13) _\n"
+ "\t\t& \"   Linux and you must install mplayer.\" & Chr (13) _\n"
+ "\t\t& \"   Relative file path is relative to the\" & Chr (13) _\n"
+ "\t\t& \"   document file.  Play once before your\" & Chr (13) _\n"
+ "\t\t& \"   live show to avoid the I/O delay with cache.\"\n"
+ "End Sub\n";
    
    /** The localization resources. */
    private ResourceBundle l10n = null;
    
    /** The image files. */
    private List<File> files = new ArrayList<File>();
    
    /**
     * The number of rows to use, default to 0 (estimate it
     * automatically).
     */
    private int numRows = 0;
    
    /** The spreadsheet document to work with. */
    private SpreadsheetDocument document = null;
    
    /** The maximum available mosaic dimension. */
    private Dimension maxDimension = null;
    
    /**
     * Creates a new instance of CalcMosaic.
     * 
     * @param args the command line arguments
     */
    public CalcMosaic(String args[]) {
        // Obtains the localization resources
        this.l10n = ResourceBundle.getBundle(
            this.getClass().getPackage().getName() + ".res.L10n");
        this.parseArguments(args);
    }
    
    /**
     * Runs CalcMosaic from the command line.
     * 
     * @param args the command line arguments
     * @throws com.sun.star.comp.helper.BootstrapException if fails to
     *         create the initial component context
     * @throws java.io.IOException when file couldn't be found or
     *         was corrupt
     */
    public static void main(String args[])
            throws com.sun.star.comp.helper.BootstrapException,
                java.io.IOException {
        CalcMosaic mosaic = new CalcMosaic(args);
        
        try {
            mosaic.create();
        } catch (com.sun.star.comp.helper.BootstrapException e) {
            throw e;
        } catch (java.io.IOException e) {
            throw e;
        }
        System.exit(0);
    }
    
    /**
     * Parses the command line arguments.
     * 
     * @param args the command line arguments
     * @throws IllegalArgumentException if the arguments are
     *         illegal
     */
    private void parseArguments(String args[])
            throws IllegalArgumentException {
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-r") || args[i].equals("-rows")) {
                i++;
                if (i >= args.length) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify the cell size.");
                    throw new IllegalArgumentException(JAR_NAME + ": "
                        + "Please specify the cell size.");
                }
                this.numRows = Integer.parseInt(args[i]);
                if (this.numRows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].startsWith("-r=")) {
                this.numRows = Integer.parseInt(args[i].substring(3));
                if (this.numRows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].startsWith("-rows=")) {
                this.numRows = Integer.parseInt(args[i].substring(6));
                if (this.numRows <= 0) {
                    System.err.println(JAR_NAME + ": "
                        + "Please specify a positive number of rows.");
                    throw new IllegalArgumentException(
                        "Please specify a positive number of rows to use.");
                }
            } else if (args[i].equals("-h") || args[i].equals("-help")) {
                System.out.println(
"Usage: java -jar " + JAR_NAME + " [options] file1 [file2...]\n"
+ "Create mosaic art from images with Calc.\n"
+ "\n"
+ "  -r,--rows num    The number of rows to use (default to\n"
+ "                   estimate it automatically).\n"
+ "  -h,--help        Display this help.\n"
+ "  -v,--version     Display version number.\n"
+ "  file1, file2...  The image files.\n"
+ "\n"
+ "The snapshots of a video can be grabbed using ffmpeg:\n"
+ "ffmpeg -i video.mp4 -r <framerate> -sameq -y -f image2 img%04d.jpg\n"
+ "Note that a spreadsheet document can only contain at most 256 spreadsheets.");
                System.exit(0);
            } else if (args[i].equals("-v") || args[i].equals("-version")) {
                System.out.println(
JAR_NAME + " v" + VERSION + " by " + AUTHOR);
                System.exit(0);
            } else if (args[i].startsWith("-")) {
                System.err.println(JAR_NAME + ": "
                    + "Unrecognized switch: " + args[i]);
                throw new IllegalArgumentException(
                    "Unrecognized switch: " + args[i]);
            } else {
                this.files.add(new File(args[i]));
            }
        }
        
        // Checks the limits of the number of images/spreadsheets.
        if (this.files.size() == 0) {
            System.err.println(JAR_NAME + ": "
                + "Please specify the image files.");
            throw new IllegalArgumentException(
                "Please specify the image files.");
        } else if (this.files.size() > 256) {
            System.err.println(JAR_NAME + ": "
                + "Too many images (256 max).  A"
                + " spreadsheet document can only contain at most 256"
                + " spreadsheets.");
            throw new IllegalArgumentException(
                "Too many images (256 max).  A spreadsheet document"
                + " can only contain at most 256 spreadsheets.");
        }
    }
    
    /**
     * Creates mosaic art of images with Calc.
     * 
     * @throws com.sun.star.comp.helper.BootstrapException if fails to
     *         create the initial component context
     * @throws java.io.IOException when file couldn't be found or was
     *         corrupt
     */
    private void create()
            throws com.sun.star.comp.helper.BootstrapException,
                java.io.IOException {
        OpenOffice office = null;
        Object param[] = new Object[1];
        Object data[][] = new Object[2][this.files.size()];
        
        try {
            office = new OpenOffice();
        } catch (com.sun.star.comp.helper.BootstrapException e) {
            throw e;
        }
        this.document = office.startNewSpreadsheetDocument();
        this.document.addBasicModule(playModule, playMacros);
        this.document.addBasicModule(basicModule, basicMacros);
        for (int i = 0; i < this.files.size(); i++) {
            data[0][i] = this.files.get(i).getName();
            try {
                data[1][i] = this.getColorMatrix(this.files.get(i));
            } catch (java.io.IOException e) {
                throw e;
            }
        }
        param[0] = data;
        this.document.runBasicMacro(
            basicModule, "subCreateMosaic", param);
    }
    
    /**
     * Calculates the mosaic color matrix from an image file.
     * 
     * @param file the source image file
     * @throws java.io.IOException if there is any problem reading
     *         the image
     */
    private Object[][] getColorMatrix(File file)
            throws java.io.IOException {
        BufferedImage image = null;
        int cellSize = -1;
        Object colors[][] = new Object[0][0];
        
        try {
            image = ImageIO.read(file);
        } catch (java.io.IOException e) {
            throw e;
        }
        
        cellSize = this.getCellSize(image);
        colors = new Object[(image.getHeight() - 1) / cellSize + 1][(image.getWidth() - 1) / cellSize + 1];
        
        // Gets the color of each cell
        for (int y0 = 0; y0 < image.getHeight(); y0 += cellSize) {
            int y1 = y0 + cellSize - 1;
            if (y1 >= image.getHeight()) {
                y1 = image.getHeight() - 1;
            }
            for (int x0 = 0; x0 < image.getWidth(); x0 += cellSize) {
                int x1 = x0 + cellSize - 1;
                if (x1 >= image.getWidth()) {
                    x1 = image.getWidth() - 1;
                }
                Color color = getMosaicColor(image, x0, y0, x1, y1);
                colors[y0 / cellSize][x0 / cellSize]
                    = color.getRGB() & 0x00FFFFFF;
            }
        }
        
        // Record the color values
        return colors;
    }
    
    /**
     * Returns the mosaic cell size.
     * 
     * @param image the source image
     * @return the calculated mosaic cell size
     */
    private int getCellSize(BufferedImage image) {
        if (this.numRows != 0) {
            return (image.getHeight() - 1) / this.numRows + 1;
        }
        
        // Calculate the size that fits the whole image.
        int width = 0, height = 0;
        
        this.maxDimension = this.getAvailableDimension();
        width = (image.getWidth() - 1) / this.maxDimension.width + 1;
        height = (image.getHeight() - 1)
            / this.maxDimension.height + 1;
        return width > height? width: height;
    }
    
    /**
     * Returns the estimated available dimension of the mosaic cells.
     * 
     * @return the estimated available dimension of the mosaic cells.
     */
    private Dimension getAvailableDimension() {
        if (this.maxDimension != null) {
            return this.maxDimension;
        }
        
        Dimension window = this.document.getWindowSize();
        int rowHeight = this.document.getRowHeight();
        
        // Only use the row height, because the height is always
        // shorter than the width in newly-created sheets.
        this.maxDimension = new Dimension(
            window.width * 25 / rowHeight,
            window.height * 25 / rowHeight);
        return this.maxDimension;
    }
    
    /**
     * Obtain the mosaic color of a block in the image, using
     * linear distance weakening.
     * 
     * @param image the image
     * @param x0 the x-coordinate of the image's upper-left corner
     * @param y0 the y-coordinate of the image's upper-left corner
     * @param x1 the x-coordinate of the image's lower-right corner
     * @param y1 the y-coordinate of the image's lower-right corner
     * @param the mosaic color of this block
     */
    private Color getMosaicColor(BufferedImage image,
            int x0, int y0, int x1, int y1) {
        double count = 0, alpha = 0, red = 0, green = 0, blue = 0;
        double xc = (double) (x0 + x1) / 2;
        double yc = (double) (y0 + y1) / 2;
        double radius = Math.sqrt((xc - x0) * (xc - x0)
            + (yc - y0) * (yc - y0)) + 1;
        
        for (int x = x0; x <= x1; x++) {
            for (int y = y0; y <= y1; y++) {
                Color color = new Color(image.getRGB(x, y));
                double distance = Math.sqrt((x - xc) * (x - xc)
                    + (y - yc) * (y - yc));
                double power = (radius - distance) / radius;
                
                alpha += color.getAlpha() * power;
                red += color.getRed() * power;
                green += color.getGreen() * power;
                blue += color.getBlue() * power;
                count += power;
            }
        }
        
        alpha /= count;
        red /= count;
        green /= count;
        blue /= count;
        return new Color((int) red, (int) green, (int) blue,
            (int) alpha);
    }
    
    /**
     * Gets a string for the given key from this resource bundle
     * or one of its parents.  If the key is missing, returns
     * the key itself.
     * 
     * @param key the key for the desired string 
     * @return the string for the given key 
     */
    private String _(String key) {
        try {
            return this.l10n.getString(key);
        } catch (java.util.MissingResourceException e) {
            return key;
        }
    }
}
